/******************************************************************************
 loadfile.pl 4/12/2003 (Pedro Cabalar)
******************************************************************************/ 

:-dynamic   rulenum/1.		% next rule identifier

rulenum(0).

loadfiles([]):-!.
loadfiles([Fname|Fs]):-
	(opt(quiet),!; header),

	atom_concat(Fname,'.tmp',TmpFile),
	
	% Remove temporal file
	(delete_file(TmpFile),!; true),
	
	preparse_files([Fname|Fs],TmpFile),

	(retractall(sortnames(_)),!;true),
	(retractall(staticnames(_)),!;true),
	asserta(sortnames([boolean])),
	asserta(staticnames([])),
	see(TmpFile), read_terms, seen,

	check_declarations,
	% Remove temporal file
	(delete_file(TmpFile),!; true).

preparse_files([],_):-!.
preparse_files([F|Fs],TmpFile):-
	(opt(quiet),!; writelist(['loading file ',F,'...']),nl),
	concatall(['getterms < ',F,' >> ',TmpFile],Command),
	shell(Command),
	preparse_files(Fs,TmpFile).

read_terms:-
	read(Term),
	(
	  Term=end_of_file,!
	   
	; ( Term=instdecls(X,L),!,
	    filter_elements(L,Vs,E),
	    add_instances(X,union(E,Vs))
	  
	  ; Term=vardecls(Vs),!,declare_var_list(Vs)
	  
	  ; Term=attrdecls(T,As,D,R,V),!,
	    (D=[],R=[],T=static,V=[X],!, 
	    	% a static fluent without domain or range
		% but with default value X
	     R1=[X]  % is assumed to have range [X]
	    ; R=[],!, R1=boolean  % if no range, assume boolean by default
	    ; R1=R),
	    declare_attrs(T,As,D,R1,V)

	  ; Term=ruledecl(H,C,Line),!,
	    treatbody(C,C1),
	    rulenum(N),incr(rulenum,1),
	    assertz(decl(rule,N/Line,[H,C1,dynamic]))
	  
	  ; Term=initdecl(H,C,Line),!,
	    treatbody(C,C1),
	    rulenum(N),incr(rulenum,1),
	    assertz(decl(rule,N/Line,[H,C1,initial]))

	  ; Term=goaldecl(H,C,Line),!,
	    treatbody(C,C1),
	    rulenum(N),incr(rulenum,1),
	    assertz(decl(rule,N/Line,[H,C1,goal]))

	  ; Term=stripsdecl(Line,H,P,E),!,
	    treatbody(P,P1),
	    assertz(decl(strips,Line,[H,P1,E]))

	  ; Term=heuristics(E),!,
	    (decl(heuristics,_,_),!,write('ERROR: multiple declarations of heuristics function'),nl,fail; true),
	    asserta(decl(heuristics,_,E))
	    
	  ; debugln(['Not implemented:',Term])
	  ),
	  read_terms
	).

treatbody(B,B):-opt(backend(_)),!.
treatbody(B,B1):-binop(true,and,B,B1).

% filter_elements(L,Vs,E)
%	L - list of set expressions set(_) mixed with element expressions
%	Vs - returns the list of all element expressions in L
%	E - returns a set union of all the set expressions in L
filter_elements([],[],[]):-!.
filter_elements([set(X)|L],Vs,union(X,E)):-
	!,filter_elements(L,Vs,E).
filter_elements([X|L],[X|Vs],E):-
	!,filter_elements(L,Vs,E).

add_instances(boolean,_):-
	write('ERROR: attempt to redefine sort "boolean".'),nl,fail.
add_instances(X,E):-
	( decl(sort,X,E0),!,retractall(decl(sort,X,_))
	; add_to_set(sortnames,X),  E0=[] % sort had not been declared yet
	), 
	asserta(decl(sort,X,union(E0,E))).

declare_var_list([]):-!.
declare_var_list([var(Vs,E)|Vars]):-declare_vars(Vs,E),declare_var_list(Vars).
	
declare_vars([],_):-!.
declare_vars([v(V)|Vs],E):-
	asserta(decl(var,v(0,V),E)),
	declare_vars(Vs,E).

declare_attrs(_T,[],_D,_R,_V):-!.
declare_attrs(T,[A|As],D,R,V):-
	length(D,N),
	( decl(attr,A/N,_), !,
	  writelist(['ERROR: multiple declarations of attribute ',A,'.']),nl,
	  fail
	; asserta(decl(attr,A/N,[T,D,R,V])),
	  ( T=static,!, add_to_set(staticnames,A/N)
	  ; true)
	
	),
	declare_attrs(T,As,D,R,V).

% Checking all the references and variables inside expressions of declarations

check_declarations:-
	bagof(decl(A,B,C),decl(A,B,C),Ds),!,
	check_expressions(Ds).
	
check_declarations.

check_expressions([]):-!.
check_expressions([decl(T,X,D)|Ds]):-

	map_subterms(checkref,[sort/1,r/3],D,D1),
	checkvars(D1,D2),
	retract(decl(T,X,D)),assertz(decl(T,X,D2)),
	
	check_expressions(Ds).


% Checks references to existing sorts and attributes (with the right arity).
% A reference like r(0,X,[]), where X is not an attribute, is replaced by
% an atomic value X.

checkref(sort(X),sort(X)):-
	!,
	( decl(sort,X,_),!
	; add_instances(X,[]),
	  writelist(['WARNING: sort ',X,' has no instances.']),nl
	).

checkref(r(0,X,[]),T):-
	!,
	( decl(attr,X/0,_),!, 	% if X is a 0-ary attribute
	  T=r(0,X,[])		% then r(0,X,[]) is ok
	; T=X			% otherwise, classify as a symbolic value
	).
	
checkref(r(P,X,L),r(P,X,L)):-
	!,
	length(L,N),
	(decl(attr,X/N,_),!
	;writelist(['ERROR: attribute ',X,'/',N,' not declared.']),nl,
	 fail
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
% Check	all variables
% A variable like v('I') is replaced by v(N,'I') where N denotes the
% depth level (0 = global var, 1 = 1st level quantified var, 2 = 2nd level, ...)
% Example
%    vars R:{1..10};
%    ...
%    forall P:{1..10} (exists Q:{1..10} (a(P)=b(Q)+R) )
%
% P is denoted as v(1,'P'), Q as v(2,'Q') and R as v(0,'R') (global var)

varname(V):-decl(var,v(0,V),_).

checkvars(T0,T):-
	findall(v(0,V),decl(var,v(0,V),_),Vs),
	checkvars(0,Vs,T0,T).
	
checkvars(_N,Vs,v(V),v(M,V)):-
	!,
	( member(v(M,V),Vs),!
	; writelist(['ERROR: variable ',V,' not declared.']),nl,fail
	).

checkvars(N,Vs,[Head0,body(C0,VDecls0),Type],[Head,body(C,VDecls2),Type]):-
	!,
	check_var_ranges(N,Vs,VDecls0,VDecls1),
	M is N+1,
	set_local_vars(M,Vs,VDecls1,Vs1,VDecls2),
	checkvars(M,Vs1,Head0,Head),
	checkvars(M,Vs1,C0,C).

checkvars(N,Vs,intensional(E0,body(C0,VDecls0)),intensional(M,E,body(C,VDecls2))):-
	!,
	check_var_ranges(N,Vs,VDecls0,VDecls1),
	M is N+1,
	set_local_vars(M,Vs,VDecls1,Vs1,VDecls2),
	checkvars(M,Vs1,E0,E),
	checkvars(M,Vs1,C0,C).

checkvars(N,Vs,T0,T):-
	!,T0 =.. [F | Args0], checkvars_args(N,Vs,Args0,Args),T  =.. [F | Args].
	
checkvars_args(_N,_Vs,[],[]):-!.
checkvars_args(N,Vs,[A|As],[B|Bs]):-
	checkvars(N,Vs,A,B),checkvars_args(N,Vs,As,Bs).

check_var_ranges(_N,_Vs,[],[]):-!.
check_var_ranges(N,Vs,[var(Us,S)|Rest],[var(Us,S1)|Rest1]):-
	checkvars(N,Vs,S,S1),check_var_ranges(N,Vs,Rest,Rest1).

set_local_vars(_N,Vs,[],Vs,[]):-!.
set_local_vars(N,Vs,[var(Us,S)|Rest],VsN,[var(Us1,S)|Rest1]):-
	bagof(v(N,U),member(v(U),Us),Us1),
	append(Us1,Vs,Vs1), % local vars are located in the head of the list
	set_local_vars(N,Vs1,Rest,VsN,Rest1).
